@@ -9,7 +9,9 @@
 #include "proto.h"
 #include "decode.h"
 
+#ifdef CAN_64BIT
 static void decode_bigint  (pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output);
+#endif
 static void decode_blob    (pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output);
 static void decode_boolean (pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output);
 static void decode_date    (pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output);
@@ -40,12 +42,6 @@ void decode_cell(pTHX_ char *input, STRLEN len, STRLEN *pos, struct cc_type *typ
     }
 
     switch (type->type_id) {
-        case CC_TYPE_BIGINT:
-        case CC_TYPE_COUNTER:
-        case CC_TYPE_TIMESTAMP:
-            decode_bigint(aTHX_ bytes, bytes_len, type, output);
-            break;
-
         case CC_TYPE_ASCII:
         case CC_TYPE_CUSTOM:
         case CC_TYPE_BLOB:
@@ -56,10 +52,6 @@ void decode_cell(pTHX_ char *input, STRLEN len, STRLEN *pos, struct cc_type *typ
             decode_boolean(aTHX_ bytes, bytes_len, type, output);
             break;
 
-        case CC_TYPE_INT:
-            decode_int(aTHX_ bytes, bytes_len, type, output);
-            break;
-
         case CC_TYPE_VARCHAR:
         case CC_TYPE_TEXT:
             decode_utf8(aTHX_ bytes, bytes_len, type, output);
@@ -91,15 +83,13 @@ void decode_cell(pTHX_ char *input, STRLEN len, STRLEN *pos, struct cc_type *typ
             decode_decimal(aTHX_ bytes, bytes_len, type, output);
             break;
 
-        case CC_TYPE_TINYINT:
-            decode_tinyint(aTHX_ bytes, bytes_len, type, output);
-            break;
-
-        case CC_TYPE_SMALLINT:
-            decode_smallint(aTHX_ bytes, bytes_len, type, output);
-            break;
-
         case CC_TYPE_VARINT:
+        case CC_TYPE_BIGINT:
+        case CC_TYPE_COUNTER:
+        case CC_TYPE_TIMESTAMP:
+        case CC_TYPE_SMALLINT:
+        case CC_TYPE_TINYINT:
+        case CC_TYPE_INT:
             decode_varint(aTHX_ bytes, bytes_len, type, output);
             break;
 
@@ -172,6 +162,7 @@ inline void bswap2(char *input)
     *the_num= ntohs(*the_num);
 }
 
+#ifdef CAN_64BIT
 void decode_bigint(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output)
 {
     union {
@@ -186,6 +177,7 @@ void decode_bigint(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *outp
     bswap8(bytes_or_bigint.bytes);
     sv_setiv(output, bytes_or_bigint.bigint);
 }
+#endif
 
 void decode_blob(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output)
 {
@@ -257,7 +249,8 @@ void decode_tinyint(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *out
     if (UNLIKELY(len != 1))
         croak("decode_tinyint: len != 1");
 
-    sv_setiv(output, *input);
+    int8_t number = *input;
+    sv_setiv(output, number);
 }
 
 void decode_utf8(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output)
@@ -385,6 +378,7 @@ void decode_varint(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *outp
         decode_int(aTHX_ bytes, 4, type, output);
     } else if (len == 4) {
         decode_int(aTHX_ input, len, type, output);
+#ifdef CAN_64BIT
     } else if (len < 8) {
         char bytes[8];
         memset(bytes, (input[0] & 0x80) ? 0xff : 0, 8);
@@ -392,6 +386,7 @@ void decode_varint(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *outp
         decode_bigint(aTHX_ bytes, 8, type, output);
     } else if (len == 8) {
         decode_bigint(aTHX_ input, len, type, output);
+#endif
     } else {
         char *tmp, *tmpout;
         struct cc_bignum bn;
@@ -421,17 +416,10 @@ void decode_varint(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *outp
     }
 }
 
-/* Perl knows how to divide and mod, but... C doesn't! So we implement our own wrappers here, to fix the behavior of C. */
-int64_t divide_properly(int64_t x, int64_t y) /* Only works for a positive 'y' */
+/* fun fact: fmod() doesn't actually implement the modulo operation... */
+double fmod_properly(double x, double y)
 {
-    int divided = x / y;
-    if (x % y != 0 && x < 0) divided--;
-    return divided;
-}
-
-int64_t mod_properly(int64_t x, int64_t y) /* Probably only works for a positive 'y' */
-{
-    int mod = x % y;
+    double mod = fmod(x, y);
     if (mod < 0)
         mod += y;
     return mod;
@@ -439,7 +427,8 @@ int64_t mod_properly(int64_t x, int64_t y) /* Probably only works for a positive
 
 void decode_date(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output)
 {
-    int64_t ind, f, e, J, h, g, Y, M, D;
+    uint32_t ind;
+    double f, e, J, h, g, Y, M, D;
 
     if (UNLIKELY(len != 4))
         croak("decode_date: len != 4");
@@ -450,17 +439,18 @@ void decode_date(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output
     J = ind;
     J -= 0x80000000 - 2440588;
 
-    f = J + 1401 + divide_properly((divide_properly((4 * J + 274277), 146097) * 3), 4) - 38;
+    f = J + 1401 + floor((floor((4 * J + 274277) / 146097) * 3) / 4) - 38;
     e = (4 * f) + 3;
-    g = divide_properly(mod_properly(e, 1461), 4);
+    g = floor(fmod_properly(e, 1461) / 4);
     h = 5 * g + 2;
-    D = divide_properly(mod_properly(h, 153), 5) + 1;
-    M = mod_properly(divide_properly(h, 153) + 2, 12) + 1;
-    Y = divide_properly(e, 1461) - 4716 + divide_properly(12 + 2 - M, 12);
+    D = floor(fmod_properly(h, 153) / 5) + 1;
+    M = fmod_properly((floor(h / 153) + 2), 12) + 1;
+    Y = floor(e / 1461) - 4716 + floor((12 + 2 - M) / 12);
 
-    sv_setpvf(output, "%lld-%.2lld-%.2lld", Y, M, D);
+    sv_setpvf(output, "%.0lf-%02.0lf-%02.0lf", Y, M, D);
 }
 
+#ifdef CAN_64BIT
 void decode_time(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output)
 {
     int64_t nano, seconds, hours, minutes;
@@ -495,6 +485,50 @@ void decode_time(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output
         pvlen--;
     SvCUR_set(output, pvlen);
 }
+#else
+//32bit compat
+void decode_time(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output)
+{
+    int32_t nano, seconds, hours, minutes;
+    char *txt;
+    char workbuf[20];
+    STRLEN txt_len;
+
+    decode_varint(aTHX_ input, len, type, output);
+    // output now contains a string represending the ns since midnight
+
+    txt = SvPV(output, txt_len);
+    if (txt_len > 14) {
+        croak("decode_time: invalid value");
+    }
+
+    if (txt_len <= 9) {
+        memset(workbuf, 0, 20);
+        memcpy(workbuf, txt, txt_len);
+        seconds = 0;
+        nano = atoi(workbuf);
+    } else {
+        memset(workbuf, 0, 20);
+        memcpy(workbuf, txt+txt_len-9, 9);
+        nano = atoi(workbuf);
+        memset(workbuf, 0, 20);
+        memcpy(workbuf, txt, txt_len-9);
+        seconds = atoi(workbuf);
+    }
+
+    hours   = seconds / 3600;
+    minutes = (seconds % 3600) / 60;
+    seconds = seconds % 60;
+
+    sv_setpvf(output, "%d:%.2d:%.2d.%d", hours, minutes, seconds, nano);
+    txt = SvPV(output, txt_len);
+    while (txt[txt_len-1] == '0')
+        txt_len--;
+    if (txt[txt_len-1] == '.')
+        txt_len--;
+    SvCUR_set(output, txt_len);
+}
+#endif
 
 void decode_map(pTHX_ char *input, STRLEN len, struct cc_type *type, SV *output)
 {
